"use strict";
// Copyright (c) Microsoft Corporation. All rights reserved. Licensed under the MIT license.
// See LICENSE in the project root for license information.
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.loadPartialTsconfigFileAsync = exports.loadTypeScriptConfigurationFileAsync = exports.PLUGIN_NAME = void 0;
const path = __importStar(require("path"));
const tapable_1 = require("tapable");
const node_core_library_1 = require("@rushstack/node-core-library");
const heft_config_file_1 = require("@rushstack/heft-config-file");
const TypeScriptBuilder_1 = require("./TypeScriptBuilder");
const anything_schema_json_1 = __importDefault(require("./schemas/anything.schema.json"));
const typescript_schema_json_1 = __importDefault(require("./schemas/typescript.schema.json"));
/**
 * The name of the plugin, as specified in heft-plugin.json
 *
 * @public
 */
exports.PLUGIN_NAME = 'typescript-plugin';
let _typeScriptConfigurationFileLoader;
const _typeScriptConfigurationFilePromiseCache = new Map();
/**
 * @beta
 */
async function loadTypeScriptConfigurationFileAsync(heftConfiguration, terminal) {
    const buildFolderPath = heftConfiguration.buildFolderPath;
    // Check the cache first
    let typescriptConfigurationFilePromise = _typeScriptConfigurationFilePromiseCache.get(buildFolderPath);
    if (!typescriptConfigurationFilePromise) {
        // Ensure that the file loader has been initialized.
        if (!_typeScriptConfigurationFileLoader) {
            _typeScriptConfigurationFileLoader = new heft_config_file_1.ConfigurationFile({
                projectRelativeFilePath: 'config/typescript.json',
                jsonSchemaObject: typescript_schema_json_1.default,
                propertyInheritance: {
                    staticAssetsToCopy: {
                        // When merging objects, arrays will be automatically appended
                        inheritanceType: heft_config_file_1.InheritanceType.merge
                    }
                }
            });
        }
        typescriptConfigurationFilePromise =
            _typeScriptConfigurationFileLoader.tryLoadConfigurationFileForProjectAsync(terminal, buildFolderPath, heftConfiguration.rigConfig);
        _typeScriptConfigurationFilePromiseCache.set(buildFolderPath, typescriptConfigurationFilePromise);
    }
    return await typescriptConfigurationFilePromise;
}
exports.loadTypeScriptConfigurationFileAsync = loadTypeScriptConfigurationFileAsync;
let _partialTsconfigFileLoader;
const _partialTsconfigFilePromiseCache = new Map();
function getTsconfigFilePath(heftConfiguration, typeScriptConfigurationJson) {
    return node_core_library_1.Path.convertToSlashes(
    // Use path.resolve because the path can start with `./` or `../`
    path.resolve(heftConfiguration.buildFolderPath, (typeScriptConfigurationJson === null || typeScriptConfigurationJson === void 0 ? void 0 : typeScriptConfigurationJson.project) || './tsconfig.json'));
}
/**
 * @beta
 */
async function loadPartialTsconfigFileAsync(heftConfiguration, terminal, typeScriptConfigurationJson) {
    const buildFolderPath = heftConfiguration.buildFolderPath;
    // Check the cache first
    let partialTsconfigFilePromise = _partialTsconfigFilePromiseCache.get(buildFolderPath);
    if (!partialTsconfigFilePromise) {
        // We don't want to load the tsconfig.json file through the rig, but we do want to take
        // advantage of the extends functionality that ConfigurationFile provides. So we'll
        // check to see if the file exists and exit early if not.
        const tsconfigFilePath = getTsconfigFilePath(heftConfiguration, typeScriptConfigurationJson);
        terminal.writeVerboseLine(`Looking for tsconfig at ${tsconfigFilePath}`);
        const tsconfigExists = await node_core_library_1.FileSystem.existsAsync(tsconfigFilePath);
        if (!tsconfigExists) {
            partialTsconfigFilePromise = Promise.resolve(undefined);
        }
        else {
            // Ensure that the file loader has been initialized.
            if (!_partialTsconfigFileLoader) {
                _partialTsconfigFileLoader = new heft_config_file_1.ConfigurationFile({
                    projectRelativeFilePath: (typeScriptConfigurationJson === null || typeScriptConfigurationJson === void 0 ? void 0 : typeScriptConfigurationJson.project) || 'tsconfig.json',
                    jsonSchemaObject: anything_schema_json_1.default,
                    propertyInheritance: {
                        compilerOptions: {
                            inheritanceType: heft_config_file_1.InheritanceType.merge
                        }
                    },
                    jsonPathMetadata: {
                        '$.compilerOptions.outDir': {
                            pathResolutionMethod: heft_config_file_1.PathResolutionMethod.resolvePathRelativeToConfigurationFile
                        }
                    }
                });
            }
            partialTsconfigFilePromise = _partialTsconfigFileLoader.loadConfigurationFileForProjectAsync(terminal, buildFolderPath, heftConfiguration.rigConfig);
        }
        _partialTsconfigFilePromiseCache.set(buildFolderPath, partialTsconfigFilePromise);
    }
    return await partialTsconfigFilePromise;
}
exports.loadPartialTsconfigFileAsync = loadPartialTsconfigFileAsync;
class TypeScriptPlugin {
    constructor() {
        this.accessor = {
            onChangedFilesHook: new tapable_1.SyncHook(['changedFilesHookOptions'])
        };
    }
    apply(taskSession, heftConfiguration) {
        taskSession.hooks.registerFileOperations.tapPromise(exports.PLUGIN_NAME, async (fileOperations) => {
            // TODO: We should consider maybe only doing one copy of static assets and pointing
            // all source files to this set of static assets. This would allow us to avoid
            // having to copy the static assets multiple times, increasing build times and
            // package size.
            for (const copyOperation of await this._getStaticAssetCopyOperationsAsync(taskSession, heftConfiguration)) {
                fileOperations.copyOperations.add(copyOperation);
            }
            return fileOperations;
        });
        taskSession.hooks.run.tapPromise(exports.PLUGIN_NAME, async (runOptions) => {
            const builder = await this._getTypeScriptBuilderAsync(taskSession, heftConfiguration);
            if (builder) {
                await builder.invokeAsync();
            }
        });
        let incrementalBuilder;
        taskSession.hooks.runIncremental.tapPromise(exports.PLUGIN_NAME, async (runIncrementalOptions) => {
            if (incrementalBuilder === undefined) {
                // eslint-disable-next-line require-atomic-updates
                incrementalBuilder = await this._getTypeScriptBuilderAsync(taskSession, heftConfiguration);
            }
            if (incrementalBuilder) {
                await incrementalBuilder.invokeAsync(runIncrementalOptions.requestRun);
            }
        });
    }
    async _getStaticAssetCopyOperationsAsync(taskSession, heftConfiguration) {
        var _a, _b, _c, _d, _e, _f;
        const typeScriptConfiguration = await loadTypeScriptConfigurationFileAsync(heftConfiguration, taskSession.logger.terminal);
        // We only care about the copy if static assets were specified.
        const copyOperations = [];
        if (((_b = (_a = typeScriptConfiguration === null || typeScriptConfiguration === void 0 ? void 0 : typeScriptConfiguration.staticAssetsToCopy) === null || _a === void 0 ? void 0 : _a.fileExtensions) === null || _b === void 0 ? void 0 : _b.length) ||
            ((_d = (_c = typeScriptConfiguration === null || typeScriptConfiguration === void 0 ? void 0 : typeScriptConfiguration.staticAssetsToCopy) === null || _c === void 0 ? void 0 : _c.includeGlobs) === null || _d === void 0 ? void 0 : _d.length) ||
            ((_f = (_e = typeScriptConfiguration === null || typeScriptConfiguration === void 0 ? void 0 : typeScriptConfiguration.staticAssetsToCopy) === null || _e === void 0 ? void 0 : _e.excludeGlobs) === null || _f === void 0 ? void 0 : _f.length)) {
            const destinationFolderPaths = new Set();
            // Add the output folder and all additional module kind output folders as destinations
            const tsconfigOutDir = await this._getTsconfigOutDirAsync(taskSession, heftConfiguration, typeScriptConfiguration);
            if (tsconfigOutDir) {
                destinationFolderPaths.add(tsconfigOutDir);
            }
            for (const emitModule of (typeScriptConfiguration === null || typeScriptConfiguration === void 0 ? void 0 : typeScriptConfiguration.additionalModuleKindsToEmit) || []) {
                destinationFolderPaths.add(`${heftConfiguration.buildFolderPath}/${emitModule.outFolderName}`);
            }
            copyOperations.push(Object.assign(Object.assign({}, typeScriptConfiguration === null || typeScriptConfiguration === void 0 ? void 0 : typeScriptConfiguration.staticAssetsToCopy), { 
                // For now - these may need to be revised later
                sourcePath: path.resolve(heftConfiguration.buildFolderPath, 'src'), destinationFolders: Array.from(destinationFolderPaths), flatten: false, hardlink: false }));
        }
        return copyOperations;
    }
    async _getTypeScriptBuilderAsync(taskSession, heftConfiguration) {
        const terminal = taskSession.logger.terminal;
        const typeScriptConfigurationJson = await loadTypeScriptConfigurationFileAsync(heftConfiguration, terminal);
        const partialTsconfigFile = await loadPartialTsconfigFileAsync(heftConfiguration, terminal, typeScriptConfigurationJson);
        if (!partialTsconfigFile) {
            // There is no tsconfig file, we can exit early
            // This check may need watch mode to break on tsconfig addition/deletion
            return false;
        }
        const typeScriptToolPath = await heftConfiguration.rigPackageResolver.resolvePackageAsync('typescript', terminal);
        // Build out the configuration
        const typeScriptBuilderConfiguration = {
            buildFolderPath: heftConfiguration.buildFolderPath,
            // Build metadata is just another build output, but we put it in the temp folder because it will
            // usually be discarded when published.
            buildMetadataFolderPath: taskSession.tempFolderPath,
            typeScriptToolPath: typeScriptToolPath,
            buildProjectReferences: typeScriptConfigurationJson === null || typeScriptConfigurationJson === void 0 ? void 0 : typeScriptConfigurationJson.buildProjectReferences,
            useTranspilerWorker: typeScriptConfigurationJson === null || typeScriptConfigurationJson === void 0 ? void 0 : typeScriptConfigurationJson.useTranspilerWorker,
            tsconfigPath: getTsconfigFilePath(heftConfiguration, typeScriptConfigurationJson),
            additionalModuleKindsToEmit: typeScriptConfigurationJson === null || typeScriptConfigurationJson === void 0 ? void 0 : typeScriptConfigurationJson.additionalModuleKindsToEmit,
            emitCjsExtensionForCommonJS: !!(typeScriptConfigurationJson === null || typeScriptConfigurationJson === void 0 ? void 0 : typeScriptConfigurationJson.emitCjsExtensionForCommonJS),
            emitMjsExtensionForESModule: !!(typeScriptConfigurationJson === null || typeScriptConfigurationJson === void 0 ? void 0 : typeScriptConfigurationJson.emitMjsExtensionForESModule),
            scopedLogger: taskSession.logger,
            emitChangedFilesCallback: (program, changedFiles) => {
                // Provide the typescript program dependent plugins
                if (this.accessor.onChangedFilesHook.isUsed()) {
                    this.accessor.onChangedFilesHook.call({ program, changedFiles });
                }
            }
        };
        // Run the builder
        const typeScriptBuilder = new TypeScriptBuilder_1.TypeScriptBuilder(typeScriptBuilderConfiguration);
        return typeScriptBuilder;
    }
    async _getTsconfigOutDirAsync(taskSession, heftConfiguration, typeScriptConfiguration) {
        var _a;
        const partialTsconfigFile = await loadPartialTsconfigFileAsync(heftConfiguration, taskSession.logger.terminal, typeScriptConfiguration);
        return (_a = partialTsconfigFile === null || partialTsconfigFile === void 0 ? void 0 : partialTsconfigFile.compilerOptions) === null || _a === void 0 ? void 0 : _a.outDir;
    }
}
exports.default = TypeScriptPlugin;
//# sourceMappingURL=TypeScriptPlugin.js.map